%
%   BW_Inverse_Kinetics.M   
%           This file provides a test case of the Inverse Kinetics Method
%     for determining Blade Worth Curves (with transient reactivity rebalancing)
%
%   The basic test scenario is as follows:
%     1.  With two given blade worth curves, we can simulate P(t) by alternately 
%         inserting the blade of interest (BOI) and then removing the compensating
%         blade (BCmp) to bring the power back up.  This sequence is repeated until  
%         the BOI is fully inserted from an initially fully withdrawn position.
%     2.  With the full P(t) vs time profile, we can use the invkin_sr routine
%         to get the full rho(t) that caused the transient.  This includes rho(t) for
%         the BOI as well as for the compensating blade.
%     3.  Use the rebank_adjust routine to isolate the rho(t) due to the BOI.
%     4.  With rho(t) and z(t) for the BOI, do a curve fit to find the integral 
%         worth curve (uses the routines directly from the umlrr_data GUI).  
%     5.  The test: If the computed integral worth curve is close to the initially 
%         assumed profile, then this should be a good test of the basic theory and 
%         actual implementation of the method
%
%   This simulation uses Blade 3 as the BOI and Blade 4 to provide reactivity 
%   compensation  --  with the blade worth data from Jan. 2018...
%
%   Let's give this a try...
%   
%   File prepared J. R. White, UMass-Lowell (last update:  March 2018)
%

      clear all,  close all,    
%%
%   define UMLRR-specific kinetics data and other parameters for the simulations
      [Be,B,Gt,lam,td,S,kappa,nu] = kinetics_data();
      tol  = 1e-5;  options = odeset('RelTol',tol); % error tolerance for ODE solver
      dt = 1.0;               % sampling time (sec)
      nlevel = 0.05;          % random noise level (random uncertainty in P(t) data)
      a = 1-nlevel;   b = 1+nlevel;  % upper and lower bounds on 'noisy' P(t)
%
      rho0 = 0;   Q = 0;      % assume critical and no source
      Po = 1;                 % initial power level (magnitude not important here)
      Co = B*Po./(lam*Gt);    % initial precursor conc
      xo = [Po Co]';          % initial condition for state vector
      rate = 4/60;            % rate of blade change (inches/sec)  (both Blade 3 & 4)
%%
%   blade locations (alternating in and out for BOI and compensating blade) 
      boi = 3;  bcmp = 4;  yr = 'Jan2018';  % BOI = Blade 3, compensation is with Blade 4 
      [dum,Hboi] = bworth(0,boi,yr);   % simply return max height of BOI
      [dum,Hcmp] = bworth(0,bcmp,yr);  % return max height of compensating blade
      zboi = [  Hboi    22   18   16   14.5   13   12    11   10    8.5     7     5   0];
      zcmp = [  4.0    6.8  9.3  10.7  12.0  13.3 14.3 15.0 16.0   17.3  18.8  20.0];
%
%   now do simulations for in & out movements for each step (two ODE calls per step)
      tend = 0;  Poo = Po;  x = xo';   rhoend = rho0;      % initialization
      rr = [];  tt = [];  PP = [];  zz3 = [];  zz4 = [];
      for i = 1:length(zboi)-1
%   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   %   do BOI insertion first   %
%   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        zlocs = zboi(i:i+1);
        dtt = (zlocs(1)-zlocs(2))/rate;  dtt = ceil(dtt)+15;  % 15 secs delay
        ts = 0:dt:dtt;  xo = x(end,:)';  rho0 = rhoend;
        ftx = @(t,x) pkeqns_rampb_nofdbk(t,x,rho0,Q,Be,Gt,B,lam,kappa,nu, ...
                                         rate,zlocs,boi,yr);
        [ts,x] = ode15s(ftx,ts,xo,options);
        tstrt = tend;  t = tstrt + ts;  tend = t(end);
        P = x(:,1);             
%   add some noise 
        rn = rand(size(ts));    % vector of uniform random numbers (0 to 1)
        srn = (b-a)*rn + a;     % uniformly distributed between a and b
        srn(1) = 1.0;           % reset initial value to 1.0
        Pn = srn.*P;            % P(t) with random noise
%   plot P(t) and z(t)
        z3 = zlocs(1)-rate*ts;  z3(z3 < zlocs(2)) = zlocs(2);  % BOI is being inserted
        z4 = zcmp(i)*ones(size(ts));
        figure(1)
        subplot(2,1,1),plot(t,Pn/Poo,'r-','LineWidth',2),grid on, hold on
        title('Inverse\_Kinetics: Simulated P(t) and Blade Position vs Time')
        ylabel('Normalized Power Level')
        subplot(2,1,2),plot(t,z3,'b--',t,z4,'g-','LineWidth',2),grid on, hold on
        xlabel('Time (seconds)'), ylabel('Blade Position (inches out)')
        legend('Blade 3','Blade 4','Location','South')
%   plot rho(t) and z(t)
        drho = bworth(z3,boi,yr)-bworth(z3(1),boi,yr);  drho = drho/100; % Dk/k
        rho = rho0+drho;  rhoend = rho(end);
        figure(2)
        subplot(2,1,1),plot(t,rho*100,'r-','LineWidth',2),grid on, hold on
        title('Inverse\_Kinetics: Simulated \rho(t) and Blade Position vs Time')
        ylabel('Reactivtiy (%\Deltak/k)')
        subplot(2,1,2),plot(t,z3,'b--',t,z4,'g-','LineWidth',2),grid on, hold on
        xlabel('Time (seconds)'), ylabel('Blade Position (inches out)')
        legend('Blade 3','Blade 4','Location','South')
%   save for subsequent processing
        if i == 1
          rr = [rr; rho];  tt = [tt; t];  PP = [PP; Pn]; 
          zz3 = [zz3 z3];  zz4 = [zz4 z4];
        else
          rr = [rr; rho(2:end)];  tt = [tt; t(2:end)];   PP = [PP; Pn(2:end)]; 
          zz3 = [zz3; z3(2:end)];  zz4 = [zz4; z4(2:end)];
        end
%   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   %   now account for the compensating blade removal   %
%   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      if i < length(zboi)-1
        zlocs = zcmp(i:i+1);
        dtt = (zlocs(2)-zlocs(1))/rate;  dtt = ceil(dtt)+15;
        ts = 0:dt:dtt;    xo = x(end,:)';  rho0 = rhoend;
        ftx = @(t,x) pkeqns_rampb_nofdbk(t,x,rho0,Q,Be,Gt,B,lam,kappa,nu, ...
                                         rate,zlocs,bcmp,yr);
        [ts,x] = ode15s(ftx,ts,xo,options);
        tstrt = tend;  t = tstrt + ts;  tend = t(end);
        P = x(:,1);             
%   add some noise
        rn = rand(size(ts));    % vector of uniform random numbers (0 to 1)
        srn = (b-a)*rn + a;     % uniformly distributed between a and b
        srn(1) = 1.0;           % reset initial value to 1.0
        Pn = srn.*P;            % P(t) with random noise
%   plot P(t) and z(t)
        z4 = zlocs(1)+rate*ts;  z4(z4 > zlocs(2)) = zlocs(2); %Bcmp is being removed
        z3 = zboi(i+1)*ones(size(ts));
        figure(1)
        subplot(2,1,1),plot(t,Pn/Poo,'r-','LineWidth',2),grid on, hold on
        title('Inverse\_Kinetics: Simulated P(t) and Blade Position vs Time')
        ylabel('Normalized Power Level')
        subplot(2,1,2),plot(t,z3,'b--',t,z4,'g-','LineWidth',2),grid on, hold on
        xlabel('Time (seconds)'), ylabel('Blade Position (inches out)')
%   plot rho(t) and z(t)
        drho = bworth(z4,bcmp,yr)-bworth(z4(1),bcmp,yr);  drho = drho/100; % Dk/k
        rho = rho0+drho;  rhoend = rho(end);
        figure(2)
        subplot(2,1,1),plot(t,rho*100,'r-','LineWidth',2),grid on, hold on
        title('Inverse\_Kinetics: Simulated \rho(t) and Blade Position vs Time')
        ylabel('Reactivtiy (%\Deltak/k)')
        subplot(2,1,2),plot(t,z3,'b--',t,z4,'g-','LineWidth',2),grid on, hold on
        xlabel('Time (seconds)'), ylabel('Blade Position (inches out)')
        legend('Blade 3','Blade 4','Location','South')
%   save for subsequent processing
        rr = [rr; rho(2:end)]; tt = [tt; t(2:end)];  PP = [PP; Pn(2:end)]; 
        zz3 = [zz3; z3(2:end)];  zz4 = [zz4; z4(2:end)];
%
      end   % end of compensating blade movement
      end   % end of forward simulation (both in and out movements)
%%
%
%   now with P(t), let's generate measured rho(t) with inverse kinetics
      rhoinvn = invkin_sr(PP,tt,lam,Co,dt,Be,Gt,B);
      nfig = 3;   figure(nfig)
      plot(tt,rhoinvn*100,'r.',tt,rr*100,'b--','LineWidth',2),grid
      title('Inverse\_Kinetics:  Comparison of Actual and Measured \rho_{tot}(t)')
      xlabel('Time (seconds)'),ylabel('Reactivity  (%\Deltak/k)')
      legend('Computed \rho(t)','Input \rho(t)','Location','SouthEast')
%
%   now do a rebank adjust and extract only the rho(t) due to the BOI
      z = [zz3 zz4];   % blade position vs time
      [rhoinv_boi] = rebank_adjust(rhoinvn,z,tt,1); % since only 2 blades (BOI here = 3)
      nfig = nfig + 1; figure(nfig)
      plot(tt,rhoinvn*100,'r.',tt,rhoinv_boi*100,'b.','LineWidth',2),grid
      title('Inverse\_Kinetics:  Total \rho(t) and BOI \rho(t)')
      xlabel('Time (seconds)'),ylabel('Reactivity  (%\Deltak/k)')
      legend('Total \rho(t)','BOI \rho(t)','Location','SouthWest')
%
%   now create blade worth curve for BOI(use routines from umlrr_data GUI)
      HHH = [Hboi Hcmp];
      [c,r2,z_FTR,rhoinv_FTR] = bwc_create(rhoinv_boi,z,1,HHH); % rho out in %Dk/k 
      Np = 250; xx = linspace(z_FTR(1),z_FTR(end),Np);    % create model data
      yy = polyval(c(1:5),xx) + c(6)*sin(2*pi*xx/Hboi); 
      totrho = yy(end)-min(yy);                           % total worth over z
      yy = yy - min(yy);                                  % makes curve positive
      rhoinv_BOIP = rhoinv_FTR - min(rhoinv_FTR);         % makes curve positive
%
%   plot integral worth curves
      rho_act = bworth(xx,boi,yr)-bworth(xx(1),boi,yr);  totrho_act = rho_act(end); 
      nfig = nfig + 1; figure(nfig)
      plot(xx,yy,'r-',xx,rho_act,'b--','Linewidth',2),grid
      title(['Inverse\_Kinetics: Integral Worth Curve -- with  r^2 = ',num2str(r2,'%6.3f')])
      xlabel('Blade Withdrawal Height (in)'),ylabel('Reactivity (%\Deltak/k)')
      legend(['Curve Fit to Simulated Data (\rho_{tot} = ',...
             num2str(totrho,'%6.3f'),' %\Deltak/k)'], ...
             ['Actual Blade 3 Worth Curve (\rho_{tot} = ',...
             num2str(totrho_act,'%6.3f'),' %\Deltak/k)'],'Location','NorthWest')
%
%   end of program
